from contextlib import contextmanager
from typing import Iterator, ContextManager, Callable

import sqlalchemy
from pydantic import BaseSettings
from sqlalchemy.orm import Session, sessionmaker
from sqlalchemy.ext.declarative import DeclarativeMeta, declarative_base


class Settings(BaseSettings):
    db_uri: str


Base: DeclarativeMeta = declarative_base()


DB_ENGINE = sqlalchemy.engine_from_config(
    {
        "url": Settings().db_uri,
        "echo": False,
    },
    prefix="",
)


def create_session() -> Iterator[Session]:
    session = sessionmaker(bind=DB_ENGINE)()

    try:
        yield session
        session.commit()
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()


db_context: Callable[[], ContextManager[Session]] = contextmanager(create_session)



from redis import Redis


# increment usage on redis
def make_key(username: str, application: str) -> str:
    return f"usage:{username}:{application}"


def record_usage(username: str, application: str, redis: Redis) -> None:
    redis.incr(make_key(username, application))



from typing import Optional, List
from pydantic import BaseModel

from fastapi import HTTPException, Depends, Request
from fastapi_jwt_auth import AuthJWT

from .schemas import UserBase


HTTP_429_TOO_MANY_REQUESTS = 429
HTTP_403_FORBIDDEN = 403


class RateLimits(BaseModel):
    limit: int
    window_secs: int


def rate_limited(key: str, limits: RateLimits, redis):
    p = redis.pipeline()
    p.incr(key)
    p.ttl(key)
    num, expire = p.execute()
    if num == 1 or expire == -1:
        redis.expire(key, limits.window_secs)
    if num > limits.limit:
        raise HTTPException(
            HTTP_429_TOO_MANY_REQUESTS,
            "Too Many Requests",
            headers={"Retry-After": str(expire)},
        )


class RateLimiter:
    def __init__(self, key: str, limits: RateLimits, redis):
        self.key = key
        self.limits = limits
        self.redis = redis

    def __call__(self, request: Request):
        if self.key == "ip":
            key = request.client.host
        else:
            key = self.key
        rate_limited(key, self.limits, self.redis)


class UserOfRole:
    def __init__(self, role: Optional[str] = None, roles: List[str] = list()):
        self.roles = [role] if role is not None else roles

    def __call__(self, Authorize: AuthJWT = Depends()):
        Authorize.jwt_required()

        roles = Authorize.get_raw_jwt().get("roles", {})
        if any(role in roles for role in self.roles):
            username = Authorize.get_jwt_subject()
            return username

        raise HTTPException(
            HTTP_403_FORBIDDEN,
            "The API key doesn't have permission to perform the request",
        )


def require_token(Authorize: AuthJWT = Depends()) -> UserBase:
    Authorize.jwt_required()
    roles = Authorize.get_raw_jwt()["roles"]
    username = Authorize.get_jwt_subject()
    return UserBase(
        username=username,
        role=roles[0],
    )


require_admin = UserOfRole(role="admin")
require_manager = UserOfRole(role="manager")
require_user = UserOfRole(role="user")
require_app = UserOfRole(role="app")
require_manager_or_admin = UserOfRole(roles=["admin", "manager"])



from datetime import datetime

from sqlalchemy import (
    Boolean,
    Column,
    Integer,
    String,
    DateTime,
    Enum,
)
from .schemas import UserType

from ..common.db_session import Base


class User(Base):
    __tablename__ = "users"

    id = Column(Integer, primary_key=True, index=True)
    username = Column(String, unique=True, index=True)
    email = Column(String, unique=True, index=True)
    salt = Column(String)
    hashed_password = Column(String)
    role = Column(Enum(UserType))
    is_active = Column(Boolean, default=True)
    created_at = Column(DateTime, default=datetime.now())



from fastapi import HTTPException, Depends
from fastapi_jwt_auth import AuthJWT


HTTP_403_FORBIDDEN = 403


def require_subscription(application: str, Authorize: AuthJWT = Depends()) -> str:
    Authorize.jwt_required()

    claims = Authorize.get_raw_jwt()
    subscription_claim = claims.get("subscription")
    if subscription_claim != application:
        raise HTTPException(
            HTTP_403_FORBIDDEN,
            "The API key doesn't have permission to perform the request",
        )
    return Authorize.get_jwt_subject()



from typing import List
from datetime import datetime

from sqlalchemy import or_
from sqlalchemy.orm.exc import NoResultFound

from ..common.queries import BaseQuery
from .models import Subscription, Usage
from .schemas import SubscriptionCreate, SubscriptionDetails, UsageDetails


class UsageException(Exception):
    pass


class SubscriptionException(Exception):
    pass


class ApplicationQuery(BaseQuery):
    def create_subscription(self, subscription_create: SubscriptionCreate):
        # check existing subscription
        found_existing_subscription = True
        try:
            self.get_active_subscription(
                subscription_create.username, subscription_create.application
            )
        except SubscriptionException:
            found_existing_subscription = False

        if found_existing_subscription:
            raise SubscriptionException(
                "Found existing subscription, please delete it before create new subscription"
            )

        new_subscription = Subscription(
            username=subscription_create.username,
            application=subscription_create.application,
            limit=subscription_create.limit,
            expires_at=subscription_create.expires_at,
            recurring=subscription_create.recurring,
            created_by=subscription_create.created_by,
        )
        self.session.add(new_subscription)
        new_usage = Usage(
            username=subscription_create.username,
            application=subscription_create.application,
            usage=0,
            starts_at=datetime.now(),
            expires_at=subscription_create.expires_at,
        )
        self.session.add(new_usage)
        self.session.commit()

    def get_active_subscription(
        self, username: str, application: str
    ) -> SubscriptionDetails:
        try:
            subscription = (
                self.session.query(Subscription)
                .filter(
                    Subscription.username == username,
                    Subscription.application == application,
                    or_(
                        Subscription.expires_at == None,  # noqa
                        Subscription.expires_at > datetime.now(),
                    ),
                )
                .one()
            )
        except NoResultFound:
            raise SubscriptionException

        return SubscriptionDetails(
            username=username,
            application=application,
            limit=subscription.limit,
            starts_at=subscription.starts_at,
            expires_at=subscription.expires_at,
            recurring=subscription.recurring,
            created_by=subscription.created_by,
            created_at=subscription.created_at,
        )

    def get_active_subscriptions(self, username: str) -> List[SubscriptionDetails]:
        try:
            subscriptions = self.session.query(Subscription).filter(
                Subscription.username == username,
                or_(
                    Subscription.expires_at is None,
                    Subscription.expires_at > datetime.now(),
                ),
            )
        except NoResultFound:
            raise SubscriptionException

        data = []
        for subscription in subscriptions:
            data.append(
                SubscriptionDetails(
                    username=subscription.username,
                    application=subscription.application,
                    limit=subscription.limit,
                    expires_at=subscription.expires_at,
                    recurring=subscription.recurring,
                )
            )
        return data

    def create_usage_from_subscription(self, subscription: SubscriptionCreate) -> Usage:
        # TODO get_or_create_usage
        # TODO try except rollback??
        new_usage = Usage(
            username=subscription.username,
            application=subscription.application,
            usage=0,
            expires_at=subscription.expires_at,
        )
        self.session.add(new_usage)
        self.session.commit()
        return new_usage

    def get_active_usage(self, username: str, application: str) -> UsageDetails:
        # TODO get all results and verify if there is only one result
        try:
            usage = (
                self.session.query(Usage)
                .filter(
                    Usage.username == username,
                    Usage.application == application,
                    or_(
                        Usage.expires_at == None,  # noqa
                        Usage.expires_at > datetime.now(),
                    ),
                )
                .one()
            )
        except NoResultFound:
            raise UsageException

        return UsageDetails(
            username=usage.username,
            application=usage.application,
            usage=usage.usage,
            starts_at=usage.starts_at,
            expires_at=usage.expires_at,
        )

    def get_active_usages(self, username: str) -> List[UsageDetails]:
        try:
            usages = self.session.query(Usage).filter(
                Usage.username == username,
                or_(
                    Usage.expires_at == None,  # noqa
                    Usage.expires_at > datetime.now(),
                ),
            )
        except NoResultFound:
            raise UsageException

        data = []
        for usage in usages:
            data.append(
                UsageDetails(
                    username=usage.username,
                    application=usage.application,
                    usage=usage.usage,
                    starts_at=usage.starts_at,
                    expires_at=usage.expires_at,
                )
            )
        return data

    def get_history_usages(self, username: str, application: str) -> List[UsageDetails]:
        usages = self.session.query(Usage).filter(
            Usage.username == username,
            Usage.application == application,
        )
        data = []
        for usage in usages:
            data.append(
                UsageDetails(
                    username=usage.username,
                    application=usage.application,
                    usage=usage.usage,
                    starts_at=usage.starts_at,
                    expires_at=usage.expires_at,
                )
            )
        return data



from datetime import datetime, timedelta
from typing import Optional

from pydantic import BaseModel, BaseSettings
from fastapi import APIRouter, Depends, HTTPException
from fastapi_jwt_auth import AuthJWT

from ..common.db_session import create_session
from ..security.schemas import UserBase  # TODO create a model for this UserBase
from ..security.depends import require_admin, require_user, require_token
from .schemas import SubscriptionCreate
from .queries import ApplicationQuery, SubscriptionException, UsageException

HTTP_429_TOO_MANY_REQUESTS = 429

router = APIRouter()


class SubscriptionSettings(BaseSettings):
    subscription_token_expires_days: int = 1


class SubscriptionIn(BaseModel):
    username: str
    application: str
    limit: int
    expires_at: Optional[datetime] = None
    recurring: bool = False


@router.post("/subscription")
def create_subscription(
    subscription: SubscriptionIn,
    username: str = Depends(require_admin),
    session=Depends(create_session),
):
    subscription_create = SubscriptionCreate(
        username=subscription.username,
        application=subscription.application,
        limit=subscription.limit,
        starts_at=datetime.now(),
        expires_at=subscription.expires_at,
        recurring=subscription.recurring,
        created_by=username,
    )
    try:
        query = ApplicationQuery(session)
        query.create_subscription(subscription_create)
        return subscription_create
    except SubscriptionException:
        return {}


@router.get("/subscription/{application}")
def get_active_subscription(
    application: str,
    user: UserBase = Depends(require_token),
    session=Depends(create_session),
):
    query = ApplicationQuery(session)
    try:
        subscription = query.get_active_subscription(user.username, application)
    except SubscriptionException:
        return {}

    return subscription


@router.get("/subscription")
def get_active_subscriptions(
    user: UserBase = Depends(require_token),
    session=Depends(create_session),
):
    if user.is_user:
        username = user.username

    query = ApplicationQuery(session)
    try:
        subscriptions = query.get_active_subscriptions(username)
    except SubscriptionException:
        return []

    return subscriptions


# delete plan
# @router.post("/subscription/_disable")
# def disable_subscription(
#     username: str = Depends(require_admin),
#     session=Depends(create_session),
# )

# disable plan

# enable plan

# get usage summary

# get usages
@router.get("/usage/{application}")
def get_active_usage(
    application: str,
    user: UserBase = Depends(require_token),
    session=Depends(create_session),
):
    query = ApplicationQuery(session)
    try:
        usage = query.get_active_usage(user.username, application)
    except UsageException:
        return {}

    return usage


@router.get("/usage")
async def get_usages(
    user: UserBase = Depends(require_token),
    session=Depends(create_session),
):
    # TODO take username for admin and manager
    query = ApplicationQuery(session)
    try:
        usages = query.get_active_usages(user.username)
    except UsageException:
        return []

    return usages


class SubscriptionTokenResponse(BaseModel):
    username: str
    application: str
    token: str
    expires_time: int


# get application token
@router.get("/token/{application}")
async def get_application_token(
    application: str,
    username: str = Depends(require_user),
    session=Depends(create_session),
):
    query = ApplicationQuery(session)
    try:
        subscription = query.get_active_subscription(username, application)
    except SubscriptionException:
        raise HTTPException(401, "NOt permitted")
    try:
        usage = query.get_active_usage(username, application)
        if usage.usage > subscription.limit:
            raise HTTPException(
                HTTP_429_TOO_MANY_REQUESTS, "You have exceeded your limit"
            )
    except UsageException:
        query.create_usage_from_subscription(subscription)

    Authorize = AuthJWT()
    expires_days = SubscriptionSettings().subscription_token_expires_days
    expires_time = timedelta(days=expires_days)
    access_token = Authorize.create_access_token(
        subject=username,
        user_claims={"subscription": application},
        expires_time=expires_time,
    )
    return SubscriptionTokenResponse(
        username=username,
        application=application,
        token=access_token,
        expires_time=expires_time.seconds,
    )



import os
import hashlib
from base64 import b64encode, b64decode


def hash_password(password, salt=None):
    if salt is None:
        salt_ = os.urandom(32)
        salt = b64encode(salt_)
    else:
        salt_ = b64decode(salt)
    hashed_password = hashlib.pbkdf2_hmac(
        "sha256",
        password.encode("utf-8"),
        salt_,
        100000,
    ).hex()
    return salt, hashed_password



from enum import Enum

from pydantic import BaseModel

from .helpers import hash_password


class UserType(str, Enum):
    USER = "user"
    MANAGER = "manager"
    APP = "app"
    ADMIN = "admin"


class UserBase(BaseModel):
    username: str
    role: UserType

    @property
    def is_admin(self) -> bool:
        return self.role == UserType.ADMIN

    @property
    def is_manager(self) -> bool:
        return self.role == UserType.MANAGER

    @property
    def is_user(self) -> bool:
        return self.role == UserType.USER

    @property
    def is_app(self) -> bool:
        return self.role == UserType.APP


class UserRegister(BaseModel):
    username: str
    email: str
    password: str


class UserCreate(UserBase):
    email: str
    password: str

    def make_user(self):
        salt, hashed_password = hash_password(self.password)
        return UserCreateHashed(
            username=self.username,
            email=self.email,
            salt=salt,
            hashed_password=hashed_password,
            role=self.role,
        )


class UserCreateHashed(UserBase):
    salt: str
    hashed_password: str


class UserSession(UserBase):
    id: int
    salt: str
    hashed_password: str


class User(UserSession):
    pass



from datetime import datetime

from sqlalchemy import (
    Boolean,
    Column,
    Integer,
    String,
    DateTime,
)

from ..common.db_session import Base


class Subscription(Base):
    __tablename__ = "subscriptions"

    id = Column(Integer, primary_key=True, index=True)
    username = Column(String, index=True)
    application = Column(String, index=True)
    limit = Column(Integer)
    starts_at = Column(DateTime, default=datetime.now())
    expires_at = Column(DateTime)
    recurring = Column(Boolean, default=False)
    created_at = Column(DateTime, default=datetime.now())
    created_by = Column(String)
    notes = Column(String)


class Usage(Base):
    __tablename__ = "usages"

    id = Column(Integer, primary_key=True, index=True)
    username = Column(String, index=True)
    application = Column(String, index=True)
    usage = Column(Integer, default=0)
    starts_at = Column(DateTime)
    expires_at = Column(DateTime)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['apihub_users',
 'apihub_users.bin',
 'apihub_users.common',
 'apihub_users.security',
 'apihub_users.subscription']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy-Utils>=0.37.4,<0.38.0',
 'SQLAlchemy>=1.4.15,<2.0.0',
 'fastapi-jwt-auth>=0.5.0,<0.6.0',
 'fastapi>=0.65.1,<0.66.0',
 'prometheus-client>=0.7.0,<0.8.0',
 'psycopg2-binary>=2.8.6,<3.0.0',
 'python-dotenv>=0.17.1,<0.18.0',
 'python-multipart>=0.0.5,<0.0.6',
 'redis>=3.5.3,<4.0.0',
 'tanbih-pipeline>=0.11.8,<0.12.0',
 'uvicorn>=0.13.4,<0.14.0']

entry_points = \
{'console_scripts': ['apihub_users_deinit = apihub_users.bin.admin:deinit',
                     'apihub_users_init = apihub_users.bin.admin:init']}

setup_kwargs = {
    'name': 'apihub-users',
    'version': '0.1.1a1',
    'description': 'user and subscription management for APIHub',
    'long_description': None,
    'author': 'Yifan Zhang',
    'author_email': 'freqyifan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import sys

from pydantic import BaseSettings

from apihub_users.common.db_session import db_context, Base, DB_ENGINE
from apihub_users.security.schemas import UserCreate, UserType
from apihub_users.security.queries import UserQuery


class SuperUser(BaseSettings):
    username: str
    password: str
    email: str

    def as_usercreate(self):
        return UserCreate(
            username=self.username,
            password=self.password,
            email=self.email,
            role=UserType.ADMIN,
        )


def init():
    Base.metadata.bind = DB_ENGINE
    Base.metadata.create_all()

    with db_context() as session:
        user = SuperUser().as_usercreate()
        UserQuery(session).create_user(user)
        sys.stderr.write(f"Admin {user.username} is created!")


def deinit():
    Base.metadata.bind = DB_ENGINE
    Base.metadata.drop_all()
    sys.stderr.write("deinit is done!")



from typing import List

from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm import Query

from ..common.queries import BaseQuery
from .models import User
from .schemas import UserSession, UserCreate
from .helpers import hash_password


class UserException(Exception):
    pass


class UserQuery(BaseQuery):
    def get_query(self) -> Query:
        return self.session.query(User)

    def get_user_by_id(self, user_id: int) -> UserSession:
        user = self.get_query().filter(User.id == user_id).one()
        return UserSession(
            id=user.id,
            username=user.username,
            role=user.role,
            salt=user.salt,
            hashed_password=user.hashed_password,
        )

    def get_user_by_username_and_password(
        self, username: str, password: str
    ) -> UserSession:
        try:
            user = self.get_query().filter(User.username == username).one()
        except NoResultFound:
            raise UserException

        _, hashed_password = hash_password(password, salt=user.salt)
        if user.hashed_password == hashed_password:
            return UserSession(
                id=user.id,
                username=user.username,
                role=user.role,
                salt=user.salt,
                hashed_password=user.hashed_password,
            )
        raise UserException

    def get_user_by_username(self, username: str) -> UserSession:
        try:
            user = self.get_query().filter(User.username == username).one()
        except NoResultFound:
            raise UserException

        return UserSession(
            id=user.id,
            username=user.username,
            role=user.role,
            salt=user.salt,
            hashed_password=user.hashed_password,
        )

    def get_users_by_usernames(self, usernames: List[str]) -> List[UserSession]:
        try:
            users = self.get_query().filter(User.username.in_(usernames))
        except NoResultFound:
            raise UserException

        return [
            UserSession(
                id=user.id,
                username=user.username,
                role=user.role,
                salt=user.salt,
                hashed_password=user.hashed_password,
            )
            for user in users
        ]

    def create_user(self, user: UserCreate) -> bool:
        """ """
        db_user = User(**user.make_user().dict())
        self.session.add(db_user)
        self.session.commit()
        # FIXME
        return True

    def change_password(self, username: str, password: str) -> bool:
        """ """
        try:
            user_in_db = self.get_query().filter(User.username == username).one()
        except NoResultFound:
            raise UserException

        _, hashed_password = hash_password(password, salt=user_in_db.salt)
        user_in_db.hashed_password = hashed_password

        return True



import datetime
from typing import List
from pydantic import BaseModel, BaseSettings
from fastapi import APIRouter, Depends, HTTPException
from fastapi.security import HTTPBasic, HTTPBasicCredentials
from fastapi_jwt_auth import AuthJWT

from ..common.db_session import create_session
from .schemas import UserCreate, UserBase, UserRegister, UserType
from .queries import UserQuery, UserException
from .depends import require_token, require_admin, require_app


security = HTTPBasic()
router = APIRouter()

HTTP_403_FORBIDDEN = 403


class SecuritySettings(BaseSettings):
    authjwt_secret_key: str = "secret"
    security_token_expires_time: int = 1


@AuthJWT.load_config
def get_config():
    return SecuritySettings()


class AuthenticateResponse(BaseModel):
    username: str
    roles: List[str]
    access_token: str
    expires_time: int


@router.get("/_authenticate")
async def _authenticate(
    credentials: HTTPBasicCredentials = Depends(security),
    session=Depends(create_session),
):
    query = UserQuery(session)
    try:
        user = query.get_user_by_username_and_password(
            username=credentials.username,
            password=credentials.password,
        )
    except UserException:
        raise HTTPException(HTTP_403_FORBIDDEN, "User not found or wrong password")

    roles = [user.role]

    Authorize = AuthJWT()
    expires_days = SecuritySettings().security_token_expires_time
    expires_time = datetime.timedelta(days=expires_days)
    access_token = Authorize.create_access_token(
        subject=user.username,
        user_claims={"roles": roles},
        expires_time=expires_time,
    )
    return AuthenticateResponse(
        username=user.username,
        roles=roles,
        expires_time=expires_time.seconds,
        access_token=access_token,
    )


@router.get("/user/{username}")
async def get_user(
    username: str,
    current_user: UserBase = Depends(require_token),
    session=Depends(create_session),
):
    if current_user.is_admin or current_user.is_manager:
        if username == "me":
            username = current_user.username
    elif username == "me":
        username = current_user.username
    else:
        raise HTTPException(HTTP_403_FORBIDDEN, "You have no permission")

    query = UserQuery(session)
    user = query.get_user_by_username(username=username)
    return UserBase(
        username=user.username,
        role=user.role,
    )


class GetUserAdminIn(BaseModel):
    usernames: str


@router.get("/user")
async def get_user_admin(
    usernames: GetUserAdminIn,
    current_username: str = Depends(require_admin),
    session=Depends(create_session),
):
    query = UserQuery(session)
    users = query.get_users_by_usernames(usernames=usernames.usernames.split(","))
    return [UserBase(**user.dict()) for user in users]


class ChangePasswordIn(BaseModel):
    password: str


@router.post("/user/_password")
async def change_password(
    password: ChangePasswordIn,
    current_user: UserBase = Depends(require_token),
    session=Depends(create_session),
):
    query = UserQuery(session)
    query.change_password(username=current_user.username, password=password.password)


@router.post("/user/{username}")
async def create_user(
    username: str,
    new_user: UserCreate,
    current_username: str = Depends(require_admin),
    session=Depends(create_session),
):
    query = UserQuery(session)
    query.create_user(new_user)
    # TODO handling results
    return {}


@router.post("/user/{username}/_password")
async def change_password_admin(
    username: str,
    password: ChangePasswordIn,
    current_username: str = Depends(require_admin),
    session=Depends(create_session),
):
    query = UserQuery(session)
    query.change_password(username=username, password=password.password)


@router.post("/register")
async def register_user(
    new_user: UserRegister,
    current_username: str = Depends(require_app),
    session=Depends(create_session),
):
    query = UserQuery(session)
    query.create_user(
        UserCreate(
            username=new_user.username,
            email=new_user.email,
            password=new_user.password,
            role=UserType.USER,
        )
    )
    return {}



from datetime import datetime
from typing import Optional

from pydantic import BaseModel


class SubscriptionBase(BaseModel):
    pass


class SubscriptionCreate(SubscriptionBase):
    username: str
    application: str
    limit: int
    starts_at: datetime = datetime.now()
    expires_at: Optional[datetime] = None
    recurring: bool = False
    created_by: str
    notes: Optional[str] = None


class SubscriptionDetails(SubscriptionCreate):
    created_at: datetime


class UsageBase(BaseModel):
    pass


class UsageCreate(UsageBase):
    username: str
    application: str
    usage: int
    starts_at: datetime
    expires_at: Optional[datetime] = None


class UsageDetails(UsageCreate):
    pass
